#!/usr/bin/env python3
# vim ts=4,fileencoding=utf-8
# SPDX-License-Identifier: Apache-2.0
# SPDXID: pymisclib-1
# PackageCopyrightText: © Copyright 2012-2022 by Christian Dönges <cd.pymisclib@platypus-projects.net >
# PackageDownloadLocation: None
# PackageHomePage: https://rot.regenbogenkraut.de/cd/pymisclib
# PackageName: pymisclib
# PackageOriginator: Originator: Platypus Projects GmbH
# PackageSourceInfo: <text>uses pymisclib from https://rot.regenbogenkraut.de/cd/pymisclib.</text>
# PackageSupplier: Christian Dönges (cd.pymisclib@platypus-projects.net )
# PackageVersion: 1.3.0
"""A queue of items sorted by the time they are due.


    :LICENSE:
    © Copyright 2017 by Christian Dönges <cd.pymisclib@platypus-projects.net >

    Licensed under the Apache License, Version 2.0 (the "License"); you may
    not use this file except in compliance with the License. You may obtain a
    copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


    If you need another license, contact the author to discuss terms.
"""

import heapq
import logging
import sys
import threading
import time


class DeadlineQueue:
    """A queue of items sorted by the time they are due.

    An item is due when the deadline expires.

    To add an element that will expire immediately, use timedue==0.

    The queue consists of tuples of the form (deadline, idx, item). The
    lowest deadline is sorted to the front of the queue.

    The deadline is based on time.perf_counter().

    The role of the idx variable is to properly order items with the same
    priority level. By keeping a constantly increasing idx, the items
    will be sorted according to the order in which they were inserted.
    """

    def __init__(self, logger=None):
        """Initialize the instance.

        :Parameters:
        logger -- The logger used by the instance. If no logger is specified,
        a default logger will be used.
        """
        # A counter used to create unique idx numbers used for sorting."""
        self._count = 0
        # A condition used to lock access to the queue.
        self._cv = threading.Condition()
        # The logger used to output information.
        if logger is None:
            self._logger = logging.getLogger(__name__)
        else:
            self._logger = logger
        # A list of items.
        self._queue = []

    def __repr__(self):
        """Return a human-readable string representation of the instance."""
        s = '['
        not_first = False
        with self._cv:
            for item in self._queue:
                if not_first:
                    s = s + f', {item[2]} @ {item[0]}'
                else:
                    not_first = True
                    s = s + f'{item[2]} @ {item[0]}'
        s = s + ']'
        return s

    def put(self, item_tuple):
        """Place a tuple in the queue.

        The item tuple is (deadline, item). A lower deadline is sorted earlier
        in the queue.
        """
        deadline = item_tuple[0]
        item = item_tuple[1]
        # Lock access to the queue by grabbing the condition.
        with self._cv:
            heapq.heappush(self._queue, (deadline, self._count, item))
            self._logger.debug(
                'put(#%u - %s @ %.6f s at %.6f s',
                self._count, item, deadline, time.perf_counter())
            self._count += 1
            # Notify anyone waiting on the condition that the queue has
            # changed.
            self._cv.notify()

    def get(self):
        """Return the first element in the queue and remove it.

        The element will be tuple (deadline, count, item).
        """
        with self._cv:
            # Wait until there is at least one element in the queue.
            while len(self._queue) == 0:
                self._cv.wait()
            element = heapq.heappop(self._queue)
        self._logger.debug(
            'get(#%u - %s @ %.6f s at %.6f s',
            element[1], element[2], element[0], time.perf_counter())
        return element

    def peek(self):
        """Returns the first element in the queue without removing it.

        The element will be tuple (deadline, count, item).
        """
        with self._cv:
            # Wait until there is at least one element in the queue.
            while len(self._queue) == 0:
                self._cv.wait()
            element = self._queue[0]
        self._logger.debug(
            'peek(#%u - %s @ %.6f s',
            element[1], element[2], element[0])
        return element

    def qsize(self):
        """Return the number of elements in the queue."""
        with self._cv:
            return len(self._queue)


if __name__ == '__main__':
    from threading import Barrier, BrokenBarrierError, Event, Thread

    start_barrier = Barrier(2, timeout=1)   # Start running producer and consumer.
    end_barrier = Barrier(3, timeout=60)    # Wait to terminate.
    terminate_event = Event()

    # Object that signals shutdown
    _sentinel = object()

    # The time offsets (in microseconds) when an event will be fired.
    event_times = [0, 0.5, 0.2, 0.1, 0, 0.01, 0.02, 0, 0.099]

    def producer(dq_out):
        """Place a new data item into the queue every interval seconds for number items."""
        start_barrier.wait()
        index = 0
        start_micros = time.perf_counter() + 0.010
        for et in event_times:
            if et != 0:
                et = start_micros + et
            # Produce some data
            data = f'Some value {index}'
            dq_out.put((et, data))
            print(f'Put: {data} @ {et:.6f} s')
            index = index + 1
        dq_out.put((0, 'abc'))
        print('Put: sentinel')
        dq_out.put((sys.float_info.max, _sentinel))
        print(f'Queue: {dq_out}')
        print(f'Producer is done, created {index} elements (plus sentinel).')
        try:
            end_barrier.wait()
        except BrokenBarrierError:
            print('producer() end_barrier broken')

    def consumer(dq_in):
        """Get data items from the queue and process them."""
        start_barrier.wait()
        items = 0
        while True:
            # Peek at the data.
            (deadline, idx, data) = dq_in.peek()
            current_s = time.perf_counter()
            if deadline > current_s:
                if terminate_event.is_set():
                    break
                continue

            # Get the first element in the queue.
            (deadline, idx, data) = dq_in.get()
            items = items + 1

            if data is _sentinel:
                print('Get: sentinel')
                break
            print(f'Get: {data} @ {deadline:.6f} s (qsize {dq_in.qsize()})'
                  f' delta_s {current_s - deadline:.6f}')

        # Indicate completion
        print(f'Consumer is done, qsize {dq_in.qsize()}.')
        print(f'Queue: {dq_in}')
        try:
            end_barrier.wait()
        except BrokenBarrierError:
            print('consumer() end_barrier broken')

    def demo():
        """Demonstration of the DeadlineQueue and how to use it."""
        # Create the shared queue and launch both threads
        q = DeadlineQueue()
        t1 = Thread(target=consumer, args=(q,))
        t2 = Thread(target=producer, args=(q,))
        t1.start()
        t2.start()
        # Wait for all produced items to be consumed
        try:
            end_barrier.wait(timeout=5)
        except BrokenBarrierError:
            print('demo() end_barrier broken')
            terminate_event.set()
        print('Both threads are done. Terminating')
        sys.exit(0)

    # Perform demonstration.
    demo()



#!/usr/bin/env python3
# vim: fileencoding=utf8
#
# SPDX-License-Identifier: Apache-2.0
# SPDXID: pymisclib-1
# PackageCopyrightText: © Copyright 2012-2022 by Christian Dönges <cd.pymisclib@platypus-projects.net >
# PackageDownloadLocation: None
# PackageHomePage: https://rot.regenbogenkraut.de/cd/pymisclib
# PackageName: pymisclib
# PackageOriginator: Originator: Platypus Projects GmbH
# PackageSourceInfo: <text>uses pymisclib from https://rot.regenbogenkraut.de/cd/pymisclib.</text>
# PackageSupplier: Christian Dönges (cd.pymisclib@platypus-projects.net )
# PackageVersion: 1.3.0
"""An interactive text console menu.

    # Example
    (pending)

    AUTHOR
    Christian Dönges, Platypus Projects GmbH www.platypus-projects.de

    The master repository for this file is located at
    https://rot.regenbogenkraut.de/cd/pymisclib.


    LICENSE
    This file is licensed under the Apache License, Version 2.0
    <https://spdx.org/licenses/Apache-2.0.html>.
    Please contact the author to discuss other licensing options.


    © Copyright 2019, 2021 Christian Dönges <cd.pymisclib@platypus-projects.net >

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging
import readline


class CommandItem:
    """An item for a command menu."""
    def __init__(self, name: str,
                 default: bool = False):
        self.name = name
        self.default = default

    def __str__(self):
        """A human-readable description of the instance."""
        s = f'CommandItem<"{self.name}"'
        if self.default:
            s += ', default>'
        else:
            s += '>'
        return s


class CommandMenu:
    """Interactive text console menu."""
    # https://pymotw.com/2/readline/
    def __init__(self,
                 vocabulary: dict = {'quit': []},
                 prompt: str = '> '):
        # Configuration members.
        self.prompt = prompt
        self.vocabulary = vocabulary

        # Dynamic members.
        self.current_candidates = []

    def __str__(self):
        """A human-readable description of the instance."""
        s = 'CommandMenu<'
        first = True
        for key, item in self.vocabulary.items():
            if first:
                s += str(item)
                first = False
            else:
                s += f', {item:!s}'
        s += '>'
        return s

    def complete(self, text, state):
        """Command-line completion of the user text in the given state."""
        response = None
        if state == 0:
            # This is the first time for this text, so build a match list.
            origline = readline.get_line_buffer()
            begin = readline.get_begidx()
            end = readline.get_endidx()
            being_completed = origline[begin:end]
            words = origline.split()

            logging.debug('origline=%s', repr(origline))
            logging.debug('begin=%s', begin)
            logging.debug('end=%s', end)
            logging.debug('being_completed=%s', being_completed)
            logging.debug('words=%s', words)

            if not words:
                self.current_candidates = sorted(self.vocabulary.keys())
            else:
                try:
                    if begin == 0:
                        # first word
                        candidates = self.vocabulary.keys()
                    else:
                        # later word
                        first = words[0]
                        candidates = self.vocabulary[first]

                    if being_completed:
                        # match vocabulary with portion of input
                        # being completed
                        self.current_candidates = \
                            [w for w in candidates if w.startswith(being_completed)]
                    else:
                        # matching empty string so use all candidates
                        self.current_candidates = candidates

                    logging.debug('candidates=%s', self.current_candidates)

                except (KeyError, IndexError) as err:
                    logging.error('completion error: %s', err)
                    self.current_candidates = []

        try:
            response = self.current_candidates[state]
        except IndexError:
            response = None
        logging.debug('complete(%s, %s) => %s', repr(text), state, response)
        return response

    def input_loop(self):
        """Loop to get input."""
        line = ''
        while line != 'quit':
            line = input(self.prompt)
            print(f'Dispatch {line}')


if __name__ == '__main__':
    # Register our completer function
    cm = CommandMenu({
        'list': ['files', 'directories'],
        'print': ['byname', 'bysize'],
        'stop': [],
    })
    readline.set_completer(cm.complete)

    # Use the tab key for completion
    readline.parse_and_bind('tab: complete')        # Works on Linux
    readline.parse_and_bind('bind ^I rl_complete')  # Works on macOS
    readline.parse_and_bind('set editing-mode vi')

    # Prompt the user for text
    cm.input_loop()



#!/usr/bin/env python3
# vim: fileencoding=utf8
#
# SPDX-License-Identifier: Apache-2.0
# SPDXID: pymisclib-1
# PackageCopyrightText: © Copyright 2023 by Christian Dönges <cd.pymisclib@platypus-projects.net >
# PackageDownloadLocation: None
# PackageHomePage: https://rot.regenbogenkraut.de/cd/pymisclib
# PackageName: pymisclib
# PackageOriginator: Originator: Platypus Projects GmbH
# PackageSourceInfo: <text>uses pymisclib from https://rot.regenbogenkraut.de/cd/pymisclib.</text>
# PackageSupplier: Christian Dönges (cd.pymisclib@platypus-projects.net )
# PackageVersion: 1.3.0
"""Format output to a text table.

    In case you have been wondering, the name is a reference to
    https://xkcd.com/327/.


    AUTHOR
    Christian Dönges, Platypus Projects GmbH www.platypus-projects.de

    The master repository for this file is located at
    https://rot.regenbogenkraut.de/cd/pymisclib.


    LICENSE
    This file is licensed under the Apache License, Version 2.0
    <https://spdx.org/licenses/Apache-2.0.html>.
    Please contact the author to discuss other licensing options.


    © Copyright 2023 Christian Dönges <cd.pymisclib@platypus-projects.net >

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import enum
from argparse import Namespace
from dataclasses import dataclass, field
from typing import Any


# Dictionary of select Unicode characters, keyed by name.
UNICODE_CHARACTERS = {
    # Unicode Box Drawing
    # Range: U+2500—U+257F
    'BOX DRAWINGS LIGHT HORIZONTAL': '─',  # U+2500
    'BOX DRAWINGS HEAVY HORIZONTAL': '━',  # U+2501
    'BOX DRAWINGS LIGHT VERTICAL': '│',  # U+2502
    'BOX DRAWINGS HEAVY VERTICAL': '┃',  # U+2503
    'BOX DRAWINGS LIGHT TRIPLE DASH HORIZONTAL': '┄',  # U+2504
    'BOX DRAWINGS HEAVY TRIPLE DASH HORIZONTAL': '┅',  # U+2505
    'BOX DRAWINGS LIGHT TRIPLE DASH VERTICAL': '┆',  # U+2506
    'BOX DRAWINGS HEAVY TRIPLE DASH VERTICAL': '┇',  # U+2507
    'BOX DRAWINGS LIGHT QUADRUPLE DASH HORIZONTAL': '┈',  # U+2508
    'BOX DRAWINGS HEAVY QUADRUPLE DASH HORIZONTAL': '┉',  # U+2509
    'BOX DRAWINGS LIGHT QUADRUPLE DASH VERTICAL': '┊',  # U+250A
    'BOX DRAWINGS HEAVY QUADRUPLE DASH VERTICAL': '┋',  # U+250B
    'BOX DRAWINGS LIGHT DOWN AND RIGHT': '┌',  # U+250C
    'BOX DRAWINGS DOWN LIGHT AND RIGHT HEAVY': '┍',  # U+250D
    'BOX DRAWINGS DOWN HEAVY AND RIGHT LIGHT': '┎',  # U+250E
    'BOX DRAWINGS HEAVY DOWN AND RIGHT': '┏',  # U+250F
    'BOX DRAWINGS LIGHT DOWN AND LEFT': '┐',  # U+2510
    'BOX DRAWINGS DOWN LIGHT AND LEFT HEAVY': '┑',  # U+2511
    'BOX DRAWINGS DOWN HEAVY AND LEFT LIGHT': '┒',  # U+2512
    'BOX DRAWINGS HEAVY DOWN AND LEFT': '┓',  # U+2513
    'BOX DRAWINGS LIGHT UP AND RIGHT': '└',  # U+2514
    'BOX DRAWINGS UP LIGHT AND RIGHT HEAVY': '┕',  # U+2515
    'BOX DRAWINGS UP HEAVY AND RIGHT LIGHT': '┖',  # U+2516
    'BOX DRAWINGS HEAVY UP AND RIGHT': '┗',  # U+2517
    'BOX DRAWINGS LIGHT UP AND LEFT': '┘',  # U+2518
    'BOX DRAWINGS UP LIGHT AND LEFT HEAVY': '┙',  # U+2519
    'BOX DRAWINGS UP HEAVY AND LEFT LIGHT': '┚',  # U+251A
    'BOX DRAWINGS HEAVY UP AND LEFT': '┛',  # U+251B
    'BOX DRAWINGS LIGHT VERTICAL AND RIGHT': '├',  # U+251C
    'BOX DRAWINGS VERTICAL LIGHT AND RIGHT HEAVY': '┝',  # U+251D
    'BOX DRAWINGS UP HEAVY AND RIGHT DOWN LIGHT': '┞',  # U+251E
    'BOX DRAWINGS DOWN HEAVY AND RIGHT UP LIGHT': '┟',  # U+251F
    'BOX DRAWINGS VERTICAL HEAVY AND RIGHT LIGHT': '┠',  # U+2520
    'BOX DRAWINGS DOWN LIGHT AND RIGHT UP HEAVY': '┡',  # U+2521
    'BOX DRAWINGS UP LIGHT AND RIGHT DOWN HEAVY': '┢',  # U+2522
    'BOX DRAWINGS HEAVY VERTICAL AND RIGHT': '┣',  # U+2523
    'BOX DRAWINGS LIGHT VERTICAL AND LEFT': '┤',  # U+2524
    'BOX DRAWINGS VERTICAL LIGHT AND LEFT HEAVY': '┥',  # U+2525
    'BOX DRAWINGS UP HEAVY AND LEFT DOWN LIGHT': '┦',  # U+2526
    'BOX DRAWINGS DOWN HEAVY AND LEFT UP LIGHT': '┧',  # U+2527
    'BOX DRAWINGS VERTICAL HEAVY AND LEFT LIGHT': '┨',  # U+2528
    'BOX DRAWINGS DOWN LIGHT AND LEFT UP HEAVY': '┩',  # U+2529
    'BOX DRAWINGS UP LIGHT AND LEFT DOWN HEAVY': '┪',  # U+252A
    'BOX DRAWINGS HEAVY VERTICAL AND LEFT': '┫',  # U+252B
    'BOX DRAWINGS LIGHT DOWN AND HORIZONTAL': '┬',  # U+252C
    'BOX DRAWINGS LEFT HEAVY AND RIGHT DOWN LIGHT': '┭',  # U+252D
    'BOX DRAWINGS RIGHT HEAVY AND LEFT DOWN LIGHT': '┮',  # U+252E
    'BOX DRAWINGS DOWN LIGHT AND HORIZONTAL HEAVY': '┯',  # U+252F
    'BOX DRAWINGS DOWN HEAVY AND HORIZONTAL LIGHT': '┰',  # U+2530
    'BOX DRAWINGS RIGHT LIGHT AND LEFT DOWN HEAVY': '┱',  # U+2531
    'BOX DRAWINGS LEFT LIGHT AND RIGHT DOWN HEAVY': '┲',  # U+2532
    'BOX DRAWINGS HEAVY DOWN AND HORIZONTAL': '┳',  # U+2533
    'BOX DRAWINGS LIGHT UP AND HORIZONTAL': '┴',  # U+2534
    'BOX DRAWINGS LEFT HEAVY AND RIGHT UP LIGHT': '┵',  # U+2535
    'BOX DRAWINGS RIGHT HEAVY AND LEFT UP LIGHT': '┶',  # U+2536
    'BOX DRAWINGS UP LIGHT AND HORIZONTAL HEAVY': '┷',  # U+2537
    'BOX DRAWINGS UP HEAVY AND HORIZONTAL LIGHT': '┸',  # U+2538
    'BOX DRAWINGS RIGHT LIGHT AND LEFT UP HEAVY': '┹',  # U+2539
    'BOX DRAWINGS LEFT LIGHT AND RIGHT UP HEAVY': '┺',  # U+253A
    'BOX DRAWINGS HEAVY UP AND HORIZONTAL': '┻',  # U+253B
    'BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL': '┼',  # U+253C
    'BOX DRAWINGS LEFT HEAVY AND RIGHT VERTICAL LIGHT': '┽',  # U+253D
    'BOX DRAWINGS RIGHT HEAVY AND LEFT VERTICAL LIGHT': '┾',  # U+253E
    'BOX DRAWINGS VERTICAL LIGHT AND HORIZONTAL HEAVY': '┿',  # U+253F
    'BOX DRAWINGS UP HEAVY AND DOWN HORIZONTAL LIGHT': '╀',  # U+2540
    'BOX DRAWINGS DOWN HEAVY AND UP HORIZONTAL LIGHT': '╁',  # U+2541
    'BOX DRAWINGS VERTICAL HEAVY AND HORIZONTAL LIGHT': '╂',  # U+2542
    'BOX DRAWINGS LEFT UP HEAVY AND RIGHT DOWN LIGHT': '╃',  # U+2543
    'BOX DRAWINGS RIGHT UP HEAVY AND LEFT DOWN LIGHT': '╄',  # U+2544
    'BOX DRAWINGS LEFT DOWN HEAVY AND RIGHT UP LIGHT': '╅',  # U+2545
    'BOX DRAWINGS RIGHT DOWN HEAVY AND LEFT UP LIGHT': '╆',  # U+2546
    'BOX DRAWINGS DOWN LIGHT AND UP HORIZONTAL HEAVY': '╇',  # U+2547
    'BOX DRAWINGS UP LIGHT AND DOWN HORIZONTAL HEAVY': '╈',  # U+2548
    'BOX DRAWINGS RIGHT LIGHT AND LEFT VERTICAL HEAVY': '╉',  # U+2549
    'BOX DRAWINGS LEFT LIGHT AND RIGHT VERTICAL HEAVY': '╊',  # U+254A
    'BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL': '╋',  # U+254B
    'BOX DRAWINGS LIGHT DOUBLE DASH HORIZONTAL': '╌',  # U+254C
    'BOX DRAWINGS HEAVY DOUBLE DASH HORIZONTAL': '╍',  # U+254D
    'BOX DRAWINGS LIGHT DOUBLE DASH VERTICAL': '╎',  # U+254E
    'BOX DRAWINGS HEAVY DOUBLE DASH VERTICAL': '╏',  # U+254F
    'BOX DRAWINGS DOUBLE HORIZONTAL': '═',  # U+2550
    'BOX DRAWINGS DOUBLE VERTICAL': '║',  # U+2551
    'BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE': '╒',  # U+2552
    'BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE': '╓',  # U+2553
    'BOX DRAWINGS DOUBLE DOWN AND RIGHT': '╔',  # U+2554
    'BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE': '╕',  # U+2555
    'BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE': '╖',  # U+2556
    'BOX DRAWINGS DOUBLE DOWN AND LEFT': '╗',  # U+2557
    'BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE': '╘',  # U+2558
    'BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE': '╙',  # U+2559
    'BOX DRAWINGS DOUBLE UP AND RIGHT': '╚',  # U+255A
    'BOX DRAWINGS UP SINGLE AND LEFT DOUBLE': '╛',  # U+255B
    'BOX DRAWINGS UP DOUBLE AND LEFT SINGLE': '╜',  # U+255C
    'BOX DRAWINGS DOUBLE UP AND LEFT': '╝',  # U+255D
    'BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE': '╞',  # U+255E
    'BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE': '╟',  # U+255F
    'BOX DRAWINGS DOUBLE VERTICAL AND RIGHT': '╠',  # U+2560
    'BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE': '╡',  # U+2561
    'BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE': '╢',  # U+2562
    'BOX DRAWINGS DOUBLE VERTICAL AND LEFT': '╣',  # U+2563
    'BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE': '╤',  # U+2564
    'BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE': '╥',  # U+2565
    'BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL': '╦',  # U+2566
    'BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE': '╧',  # U+2567
    'BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE': '╨',  # U+2568
    'BOX DRAWINGS DOUBLE UP AND HORIZONTAL': '╩',  # U+2569
    'BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE': '╪',  # U+256A
    'BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE': '╫',  # U+256B
    'BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL': '╬',  # U+256C
    'BOX DRAWINGS LIGHT ARC DOWN AND RIGHT': '╭',  # U+256D
    'BOX DRAWINGS LIGHT ARC DOWN AND LEFT': '╮',  # U+256E
    'BOX DRAWINGS LIGHT ARC UP AND LEFT': '╯',  # U+256F
    'BOX DRAWINGS LIGHT ARC UP AND RIGHT': '╰',  # U+2570
    'BOX DRAWINGS LIGHT DIAGONAL UPPER RIGHT TO LOWER LEFT': '╱',  # U+2571
    'BOX DRAWINGS LIGHT DIAGONAL UPPER LEFT TO LOWER RIGHT': '╲',  # U+2572
    'BOX DRAWINGS LIGHT DIAGONAL CROSS': '╳',  # U+2573
    'BOX DRAWINGS LIGHT LEFT': '╴',  # U+2574
    'BOX DRAWINGS LIGHT UP': '╵',  # U+2575
    'BOX DRAWINGS LIGHT RIGHT': '╶',  # U+2576
    'BOX DRAWINGS LIGHT DOWN': '╷',  # U+2577
    'BOX DRAWINGS HEAVY LEFT': '╸',  # U+2578
    'BOX DRAWINGS HEAVY UP': '╹',  # U+2579
    'BOX DRAWINGS HEAVY RIGHT': '╺',  # U+257A
    'BOX DRAWINGS HEAVY DOWN': '╻',  # U+257B
    'BOX DRAWINGS LIGHT LEFT AND HEAVY RIGHT': '╼',  # U+257C
    'BOX DRAWINGS LIGHT UP AND HEAVY DOWN': '╽',  # U+257D
    'BOX DRAWINGS HEAVY LEFT AND LIGHT RIGHT': '╾',  # U+257E
    'BOX DRAWINGS HEAVY UP AND LIGHT DOWN': '╿',  # U+257F
}  # UNICODE_CHARACTERS


# Box drawing characters.
# The beauty of Namespace makes it possible to access a character like so:
# box_drawings.light.horizontal.
box_drawings = Namespace(
    double=Namespace(
        down_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL'],
        down_and_left=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE DOWN AND LEFT'],
        down_and_right=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE DOWN AND RIGHT'],
        horizontal=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE HORIZONTAL'],
        up_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE UP AND HORIZONTAL'],
        up_and_left=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE UP AND LEFT'],
        up_and_right=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE UP AND RIGHT'],
        vertical=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE VERTICAL'],
        vertical_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL'],
        vertical_and_left=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE VERTICAL AND LEFT'],
        vertical_and_right=UNICODE_CHARACTERS['BOX DRAWINGS DOUBLE VERTICAL AND RIGHT'],
    ),  # double
    light=Namespace(
        down_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT DOWN AND HORIZONTAL'],
        down_and_left=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT DOWN AND LEFT'],
        down_and_right=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT DOWN AND RIGHT'],
        horizontal=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT HORIZONTAL'],
        up_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT UP AND HORIZONTAL'],
        up_and_left=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT UP AND LEFT'],
        up_and_right=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT UP AND RIGHT'],
        vertical=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT VERTICAL'],
        vertical_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL'],
        vertical_and_left=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT VERTICAL AND LEFT'],
        vertical_and_right=UNICODE_CHARACTERS['BOX DRAWINGS LIGHT VERTICAL AND RIGHT'],
    ),  # light
    heavy=Namespace(
        down_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY DOWN AND HORIZONTAL'],
        down_and_left=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY DOWN AND LEFT'],
        down_and_right=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY DOWN AND RIGHT'],
        horizontal=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY HORIZONTAL'],
        up_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY UP AND HORIZONTAL'],
        up_and_left=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY UP AND LEFT'],
        up_and_right=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY UP AND RIGHT'],
        vertical=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY VERTICAL'],
        vertical_and_horizontal=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL'],
        vertical_and_left=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY VERTICAL AND LEFT'],
        vertical_and_right=UNICODE_CHARACTERS['BOX DRAWINGS HEAVY VERTICAL AND RIGHT'],
    )  # heavy
)  # box_drawings


@enum.unique
class LineStyle(enum.Enum):
    """Rendering style for lines."""
    double = box_drawings.double
    light = box_drawings.light
    heavy = box_drawings.heavy


Text = list[str]


@dataclass
class ColumnTable:
    """Table with data arranged by column."""
    _headings: list[str] = field(default_factory=list)  # row or column heading
    _formats: list[str] = field(default_factory=list)  # cell formats
    _cells: list[list[Any]] = field(default_factory=list)  # column[row]
    _num_columns: int = 0
    _num_rows: int = 0

    @property
    def cell_formats(self) -> list[str]:
        """Return headings."""
        return self._formats

    @property
    def headings(self) -> list[str]:
        """Return headings."""
        return self._headings

    @property
    def num_columns(self):
        return self._num_columns

    @property
    def num_rows(self):
        return self._num_rows

    def set_table(self,
                  headings: list[str],
                  cell_formats: list[str],
                  cells: list[list[str]]):
        """Set the headings, the format of the cells, and the cell content.

           The cells are specified column[row].
        """
        if len(headings) != len(cell_formats):
            raise ValueError('number columns in headings and formats do not match')
        if len(headings) != len(cells):
            raise ValueError('number of columns in headings and cells do not match')
        self._num_columns = len(headings)
        self._num_rows = len(cells[0])
        self._headings = headings
        self._formats = cell_formats
        self._cells = cells

    def set_table_transposed(self,
                             headings: list[str],
                             cell_formats: list[str],
                             cells: list[list[str]]):
        """The cells are specified row[column]."""
        transposed_cells = [list(x) for x in zip(*cells)]
        self.set_table(headings, cell_formats, transposed_cells)

    def draw(self, style: LineStyle) -> Text:
        """Draw the table to a list of lines."""
        formatted_cells = []
        column_widths = []
        for c in range(self.num_columns):
            column_width = len(self._headings[c])
            column_data = self._cells[c]
            fmt = self._formats[c]
            formatted_column = []
            for cell in column_data:
                formatted_cell = f'{cell:{fmt}}'
                formatted_column.append(formatted_cell)
                column_width = max(column_width,  len(formatted_cell))
            formatted_cells.append(formatted_column)
            column_widths.append(column_width)

        box = style.value
        lines = []
        first = True
        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            if first:
                s = box.down_and_right + box.horizontal * (column_width + 2)
                first = False
            else:
                s += box.down_and_horizontal + box.horizontal * (column_width + 2)
        s += box.down_and_left
        lines.append(s)

        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            s += box.vertical + f' {self._headings[c]:{column_width}s} '
        s += box.vertical
        lines.append(s)

        first = True
        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            if first:
                s = box.vertical_and_right + box.horizontal * (column_width + 2)
                first = False
            else:
                s += box.vertical_and_horizontal + box.horizontal * (column_width + 2)
        s += box.vertical_and_left
        lines.append(s)

        for r in range(self.num_rows):
            s = ''
            for c in range(self.num_columns):
                column_width = column_widths[c]
                s += box.vertical + f' {formatted_cells[c][r]:{column_width}s} '
            s += box.vertical
            lines.append(s)

        first = True
        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            if first:
                s = box.up_and_right + box.horizontal * (column_width + 2)
                first = False
            else:
                s += box.up_and_horizontal + box.horizontal * (column_width + 2)
        s += box.up_and_left
        lines.append(s)

        return lines


def draw_box(x, y, style) -> Text:
    """Construct a box of size x * y with style."""
    box = style.value

    lines = []
    s = box.down_and_right + box.horizontal * x + box.down_and_left
    lines.append(s)
    s = box.vertical + ' ' * x + box.vertical
    for i in range(y):
        lines.append(s)
    s = box.up_and_right + box.horizontal * x + box.up_and_left
    lines.append(s)

    return lines


def print_lines(lines: Text):
    """Print lines stored as a list of strings to stdout."""
    for line in lines:
        print(line)


if __name__ == '__main__':
    print_lines(draw_box(5, 3, LineStyle.light))
    print_lines(draw_box(1, 1, LineStyle.double))
    print_lines(draw_box(40, 3, LineStyle.heavy))

    ct1 = ColumnTable()
    ct1.set_table_transposed(
        headings=['First', 'Second', 'Third', 'Fourth'],
        cell_formats=['3d', '08x', 's', '>s'],
        cells=[[1, 0x12345678, 'abc def geh', 'this is a sample text'],
               [2, 0xffffee01, '12345', 'Short'],
               [12, -1, 'minus one', 'negative hex number'],
               [123, 0, 'zero', 'Zero hexadecimal number'],
               [1000, 23, 'First', 'The first column is too large.']]
    )
    print_lines(ct1.draw(LineStyle.light))



#!/usr/bin/env python3
# vim ts=4,fileencoding=utf-8
# SPDX-License-Identifier: Apache-2.0
# SPDXID: pymisclib-1
# PackageCopyrightText: © Copyright 2012-2022 by Christian Dönges <cd.pymisclib@platypus-projects.net >
# PackageDownloadLocation: None
# PackageHomePage: https://rot.regenbogenkraut.de/cd/pymisclib
# PackageName: pymisclib
# PackageOriginator: Originator: Platypus Projects GmbH
# PackageSourceInfo: <text>uses pymisclib from https://rot.regenbogenkraut.de/cd/pymisclib.</text>
# PackageSupplier: Christian Dönges (cd.pymisclib@platypus-projects.net )
# PackageVersion: 1.3.0
"""A socket used to send and receive message chunks over a TCP connection.


    :LICENSE:
    © Copyright 2017, 2022 by Christian Dönges <cd.pymisclib@platypus-projects.net >

    Licensed under the Apache License, Version 2.0 (the "License"); you may
    not use this file except in compliance with the License. You may obtain a
    copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


    If you need another license, contact the author to discuss terms.
"""

# pylint: disable=consider-using-assignment-expr

from dataclasses import dataclass
import logging
import socket


@dataclass
class ChunkyStreamSocket:
    """Socket used to send message chunks over a TCP connection."""
    host: str = '127.0.0.1'
    port: int = 10000
    sock: socket.socket = None
    logger: logging.Logger = logging.getLogger(__name__)
    debug: bool = False
    log_n_bytes: int = 0   # number of communicated bytes to log
    _backlog_bytes: bytes = None

    def __post_init__(self):
        """Custom initializer called after __init__().
        """
        if self.sock is None:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def bind_and_listen(self, backlog: int = 5, timeout: float = None):
        """Bind to the host and port to make a server socket.

            : param int backlog: The maximum number of queued connections.
            : param float timeout: The number of seconds after which socket
                 operations will time out. Set to None for a blocking socket.
        """
        self.logger.debug('bind_and_listen(%s:%u)', self.host, self.port)
        # Set the timeout.
        self.sock.settimeout(timeout)
        # Allow the server socket to re-bind immediately.
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # Bind the socket to the given address and port.
        self.sock.bind((self.host, self.port))
        # Start listening for a connection attempt.
        self.sock.listen(backlog)

    def accept(self) -> tuple[int, tuple[str, int]]:
        """Accept a client connection on the (server) socket.

           :return: The client socket and client address tuple.
           :rtype tuple(socket, tuple(host, port))
           :raises: TimeoutError
        """
        try:
            (client_sock, client_addr) = self.sock.accept()  # pylint: disable=no-member
        except socket.timeout as ex:
            raise TimeoutError('accept() timed out') from ex
        self.logger.debug('accept(%s)', client_addr)
        return (client_sock, client_addr)

    def close(self):
        """Close the socket."""
        try:
            self.sock.shutdown(socket.SHUT_RDWR)
        except OSError:
            # Ignore errors in shutdown since we close the socket anyways.
            pass
        self.sock.close()
        self.sock = None
        self.logger.debug('close()')

    def connect(self, timeout: float = None):
        """Connect client socket to the server at host:port.

        :param float timeout: The number of seconds after which socket
            operations will time out. Set to None for a blocking socket.
        :raises: ConnectionError
        """
        self.logger.debug('connect(%s:%u)', self.host, self.port)
        # Set the timeout.
        self.sock.settimeout(timeout)
        # Allow server socket to re-bind immediately.
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            self.sock.connect((self.host, self.port))
        except OSError as ex:
            self.logger.error('Connection attempt to %s:%u failed: %s',
                              self.host, self.port, ex)
            raise ConnectionError('unable to connect') from ex
        self.logger.debug('Connected to %s:%s', self.host, self.port)

    def send(self, msg_bytes: bytes) -> int:
        """Send the bytes.

        :Parameters:
        msg_bytes -- The bytes to send. If the receiver is expecting a
        separator, it must be appended to the message by the caller.

        :Return:
        The number of bytes sent.

        :Raises:
        RuntimeError
        """
        total_nr_sent = 0
        while total_nr_sent < len(msg_bytes):
            current_nr_sent = self.sock.send(msg_bytes[total_nr_sent:])
            if current_nr_sent == 0:
                self.logger.debug('self.sock.send() failed.')
                raise RuntimeError("socket connection broken")
            total_nr_sent = total_nr_sent + current_nr_sent
        if self.log_n_bytes != 0:
            self.logger.debug('--> %s', msg_bytes[:self.log_n_bytes])
        return total_nr_sent

    def recv(self, length: int, timeout: float = None):
        """Receive length bytes from the socket.

        This function handles chunked data (i.e. the data to receive is split
        into multiple packets). If more data than expected is received, it is
        placed into a backlog buffer until the next call to a receive
        function.

        :Parameters:
        length -- The number of bytes to read.
        timeout -- Timeout in seconds or None to block.

        :Return:
        The received bytes.

        :Raises:
        RuntimeError
        TimeoutError
        """
        chunks = []
        nr_received = 0
        # Retrieve the backlog from the previous recv() call and use it as the
        # first chunk.
        if self._backlog_bytes is not None:
            chunks.append(self._backlog_bytes)
            nr_received = len(self._backlog_bytes)
            self._backlog_bytes = None

        # Set the timeout.
        self.sock.settimeout(timeout)

        # Receive bytes until we have enough to satisfy the length requirement.
        while nr_received < length:
            recv_len = min(length - nr_received, 4096)
            chunk = self.sock.recv(recv_len)
            if self.debug:
                self.logger.debug('socket.recv(%u) := %s', recv_len, chunk)
            if chunk == b'':
                # Connection was closed.
                break
            chunks.append(chunk)
            nr_received = nr_received + len(chunk)

        # Join all chunks into one message.
        msg_bytes = b''.join(chunks)

        # Check if the connection was closed.
        if len(msg_bytes) == 0:
            self.logger.debug('No data received, connection closed.')
            raise RuntimeError("socket connection broken")

        # Cut off the part that is too long.
        if len(msg_bytes) > length:
            self._backlog_bytes = msg_bytes[length:]
            msg_bytes = msg_bytes[:length]

        if self.log_n_bytes != 0:
            self.logger.debug('<-- %s', msg_bytes[:self.log_n_bytes])
        return msg_bytes

    def recv_to_separator(self, separator: bytes):
        """Receive bytes until the given separator is found.

        This function handles chunked data (i.e. the data to receive is
        split into multiple packets). If more data than expected is
        received, it is placed into a backlog buffer until the next call
        to a receive function.

        :Parameters:
        separator -- One or more bytes that separate messages in the TCP
        stream.

        :Return:
        The received bytes.

        :Raises:
        RuntimeError
        TimeoutError
        """
        self.logger.debug('recv_to_separator(%s)', separator)
        start_search_index = 0
        chunk = b''
        msg_bytes = b''
        while True:
            if self._backlog_bytes is not None and len(self._backlog_bytes) > 0:
                # The first time around, process the backlog.
                chunk = self._backlog_bytes
                self._backlog_bytes = None
                if self.debug:
                    self.logger.debug('backlog chunk = %s', chunk)
            else:
                chunk = self.sock.recv(4096)
                if self.debug:
                    self.logger.debug('socket.recv(4096) := %s', chunk)
            if chunk == b'':
                raise RuntimeError("socket connection broken")

            msg_bytes = msg_bytes + chunk
            start_separator_index = msg_bytes.find(separator, start_search_index)
            if start_separator_index > -1:
                # We found the separator at index start_separator_index.
                self._backlog_bytes = msg_bytes[start_separator_index + len(separator):]
                if self.debug:
                    self.logger.debug('Backlog: %u bytes', self._backlog_bytes)
                msg_bytes = msg_bytes[:start_separator_index]
                break
            # The separator could have started in the current chunk but
            # finishes in the next chunk, so we need to search the
            # len(separator) - 1 last bytes of the separator again
            start_search_index = max(0, len(msg_bytes) - (len(separator) - 1))

        if self.log_n_bytes != 0:
            self.logger.debug('<-- %s', msg_bytes[:self.log_n_bytes])
        return msg_bytes


if __name__ == '__main__':
    import sys
    from threading import Barrier, BrokenBarrierError, Thread
    from time import sleep

    start_barrier = Barrier(2, timeout=5)
    end_barrier = Barrier(3, timeout=60)
    MSG_SEP = b'\0\1\2\1\0'

    messages = [
        b'abcdef',
        b'1234',
        b'a', b'bc', b'de\0', b'\1\2\1\0fghi',
        b'xyzZYX',
        b'',
        b'+++',
    ]

    def server():
        """Test server."""
        logger = logging.getLogger('server')
        logger.info('server starting')
        server_socket = ChunkyStreamSocket(logger=logging.getLogger('server.socket'))
        server_socket.bind_and_listen(timeout=60)
        logger.debug('server_socket = %s', server_socket.sock)
        start_barrier.wait()

        logger.info('server running')
        (accepted_socket, client_addr) = server_socket.accept()
        logger.debug('Accepted client from %s:%u', client_addr[0], client_addr[1])
        cs = ChunkyStreamSocket(sock=accepted_socket,
                                logger=logging.getLogger('server.cs'))
        while True:
            msg = cs.recv_to_separator(MSG_SEP)
            logger.info('MSG: %s', msg)
            if msg == b'+++':
                logger.debug('EOF received')
                break

        logger.info('server closing connection')
        cs.close()
        server_socket.close()
        try:
            end_barrier.wait()
        except BrokenBarrierError:
            logger.error('server() end_barrier broken')

    def client():
        """Test client."""
        logger = logging.getLogger('client')
        logger.info('client starting')
        client_socket = ChunkyStreamSocket(logger=logging.getLogger('client.socket'))
        logger.debug('client_socket = %s', client_socket.sock)
        start_barrier.wait()
        logger.info('client running')
        client_socket.connect()
        for message in messages:
            try:
                client_socket.send(message + MSG_SEP)
            except RuntimeError as e:
                logger.critical('send(%u) failed: %s', len(message + MSG_SEP), e)
                logger.critical('client terminating')
                break
        try:
            end_barrier.wait()
        except BrokenBarrierError:
            logger.error('client() end_barrier broken')
        finally:
            client_socket.close()

    # Log everything to the console.
    main_logger = logging.getLogger()
    main_logger.setLevel(logging.NOTSET)
    ch = logging.StreamHandler()
    ch.setLevel(logging.NOTSET)
    formatter = logging.Formatter('%(asctime)s - %(name)20s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    main_logger.addHandler(ch)

    server_thread = Thread(target=server, args=())
    client_thread = Thread(target=client, args=())
    server_thread.start()
    sleep(1.0)  # give the server time to set up before starting client.
    client_thread.start()
    main_logger.info('Client and server running.')
    try:
        end_barrier.wait(timeout=5)
    except BrokenBarrierError:
        main_logger.error('Barrier broken. Terminating')
    sys.exit(0)



#!/usr/bin/env python3
# vim ts=4,fileencoding=utf-8
# SPDX-License-Identifier: Apache-2.0
# SPDXID: pymisclib-1
# PackageCopyrightText: © Copyright 2012-2022 by Christian Dönges <cd.pymisclib@platypus-projects.net >
# PackageDownloadLocation: None
# PackageHomePage: https://rot.regenbogenkraut.de/cd/pymisclib
# PackageName: pymisclib
# PackageOriginator: Originator: Platypus Projects GmbH
# PackageSourceInfo: <text>uses pymisclib from https://rot.regenbogenkraut.de/cd/pymisclib.</text>
# PackageSupplier: Christian Dönges (cd.pymisclib@platypus-projects.net )
# PackageVersion: 1.3.0

"""Output an ASCII table in various formats.

    :LICENSE:
    © Copyright 2019 by Christian Dönges <cd.pymisclib@platypus-projects.net >

    Licensed under the Apache License, Version 2.0 (the "License"); you may
    not use this file except in compliance with the License. You may obtain a
    copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


    If you need another license, contact the author to discuss terms.
"""

# pylint: disable=consider-using-assignment-expr

import argparse


ascii_names = {
    0: 'NUL',
    1: 'SOH',
    2: 'STX',
    3: 'ETX',
    4: 'EOT',
    5: 'ENQ',
    6: 'ACK',
    7: 'BEL',
    8: 'BS',
    9: 'HT',
    10: 'LF',
    11: 'VT',
    12: 'FF',
    13: 'CR',
    14: 'SO',
    15: 'SI',
    16: 'DC0',
    17: 'DC1',
    18: 'DC2',
    19: 'DC3',
    20: 'DC4',
    21: 'NAK',
    22: 'SYN',
    23: 'ETB',
    24: 'CAN',
    25: 'EM',
    26: 'SUB',
    27: 'ESC',
    28: 'FS',
    29: 'GS',
    30: 'RS',
    31: 'US',
    127: 'DEL',
}


def text_list():
    """Output ASCII table as Python list."""
    print('ascii_list = [')
    for i in range(256):
        if i in ascii_names:
            c = ascii_names[i]
        else:
            c = str(chr(i))
            if not c.isprintable():
                c = '_'
        if c == "'":
            print(f'    "{c}",  # {i} / 0x{i:02x} / {c}')
        elif c == '\\':
            print(f"    '\\',  # {i} / 0x{i:02x} / BACKSLASH")
        else:
            print(f"    '{c}',  # {i} / 0x{i:02x} / {c}")
    print(']')


def text_table():
    """Output ASCII table as text."""
    h1 = '     |'
    h2 = '=====+'
    for lower_nibble in range(0, 16):
        h1 += f'| {lower_nibble:04b} '
        h2 += '+======'
    body = []
    for upper_nibble in range(0, 16):
        s1 = '     |'
        s2 = f'{upper_nibble:04b} |'
        s3 = '     |'
        s4 = '-----+'
        for lower_nibble in range(0, 16):
            i = (upper_nibble << 4) + lower_nibble
            if i in ascii_names:
                c = ascii_names[i]
            else:
                c = str(chr(i))
                if not c.isprintable():
                    c = '???'
            s1 += f'| {i:#04x} '
            s2 += f'|  {i:3d} '
            s3 += f'| {c:4s} '
            s4 += '+------'
        body.append(s1)
        body.append(s2)
        body.append(s3)
        if upper_nibble < 15:
            body.append(s4)

    print(h1)
    print(h2)
    for line in body:
        print(line)


def html_list():
    """Print an ASCII Python 'list' in HTML format."""
    h1 = '<tr>'
    for lower_nibble in range(0, 16):
        h1 += f'<th>{lower_nibble:04b}</th>'
    h1 += '</tr>'
    body = []
    for upper_nibble in range(0, 16):
        s = f'<tr><th>{upper_nibble:04b}</th>'
        for lower_nibble in range(0, 16):
            i = (upper_nibble << 4) + lower_nibble
            c = ascii_names.get(i, f'&#x{i:02x};')
            s += f'<td>{i:#04x}</br>{i:3d}</br>{c}</td>'
        s += '</tr>'
        body.append(s)

    print('''<!DOCTYPE html>
<html>
<head>
    <title>ASCII Table</title>
</head>
<style>
table {
  border-collapse: collapse;
}
table, th, td {
    border: 1px solid black;
    padding: 8px;
    text-align: center;
}
th {
    border: 2px solid;
    vertical-align: middle;
}
tr {
    vertical-align: top;
}
</style>
<body>
    <h1>ASCII Table</h1>
    <pre>
''')
    print('ascii_list = [')
    for i in range(256):
        if i in ascii_names:
            c = ascii_names[i]
        else:
            c = str(chr(i))
            if not c.isprintable():
                c = f'&#x{i:02x};'
        if c == "'":
            print(f'    "{c}",  # {i} / 0x{i:02x} / {c}\r')
        elif c == '\\':
            print(f"    '\\',  # {i} / 0x{i:02x} / BACKSLASH\r")
        else:
            print(f"    '{c}',  # {i} / 0x{i:02x} / {c}\r")
    print(']')

    print('''
    </pre>
</body>
</html>
''')


def html_table():
    """Print an ASCII table in HTML format."""
    h1 = '<tr>'
    for lower_nibble in range(0, 16):
        h1 += f'<th>{lower_nibble:04b}</th>'
    h1 += '</tr>'
    body = []
    for upper_nibble in range(0, 16):
        s = f'<tr><th>{upper_nibble:04b}</th>'
        for lower_nibble in range(0, 16):
            i = (upper_nibble << 4) + lower_nibble
            c = ascii_names.get(i, f'&#x{i:02x};')
            s += f'<td>{i:#04x}</br>{i:3d}</br>{c}</td>'
        s += '</tr>'
        body.append(s)

    print('''<!DOCTYPE html>
<html>
<head>
    <title>ASCII Table</title>
</head>
<style>
table {
  border-collapse: collapse;
}
table, th, td {
    border: 1px solid black;
    padding: 8px;
    text-align: center;
}
th {
    border: 2px solid;
    vertical-align: middle;
}
tr {
    vertical-align: top;
}
</style>
<body>
    <h1>ASCII Table</h1>
    <table>
    <tr><th rowspan="2">Upper Nibble</th><th colspan="16">Lower Nibble</th></tr>
''')
    print(h1)
    for line in body:
        print(line)
    print('</table>')
    print('''
</body>
</html>
''')


def main():
    """It all happens here."""
    parser = argparse.ArgumentParser(description='Output an ASCII table.')
    parser.add_argument(
        '-f', '--format', default='text', type=str,
        choices=['html', 'list', 'python', 'text'],
        help='Output format for the table.')
    args = parser.parse_args()

    if args.format == 'html':
        html_table()
    elif args.format == 'list':
        html_list()
    elif args.format == 'python':
        text_list()
    else:
        text_table()


# This code is executed if the file is called as a stand-alone command.
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# vim: fileencoding=utf8
#
# SPDX-License-Identifier: Apache-2.0
# SPDXID: pymisclib-1
# PackageCopyrightText: © Copyright 2012-2022 by Christian Dönges <cd.pymisclib@platypus-projects.net >
# PackageDownloadLocation: None
# PackageHomePage: https://rot.regenbogenkraut.de/cd/pymisclib
# PackageName: pymisclib
# PackageOriginator: Originator: Platypus Projects GmbH
# PackageSourceInfo: <text>uses pymisclib from https://rot.regenbogenkraut.de/cd/pymisclib.</text>
# PackageSupplier: Christian Dönges (cd.pymisclib@platypus-projects.net )
# PackageVersion: 1.3.0
"""Print multiple progress bars to the console.

    The configuration of the bars is handled through a list of dictionaries.
    Each dictionary contains the parameters for a single progress bar.

    Before the bars are shown, prepare_bars() must be called to make room for
    the bars on the console.

    Each time a bar is updated, print_bars() is called to update the console
    output.

    # Bar example
    ```
    Total |------------------------------| 0% Writing
          |███████████████---------------| 50.9% /tmp3muil9sc/0000000000000000.tmp
    ```

    from multiprogressbar import prepare_bars, print_bars

    config = [
        { 'current_iteration': 0,
          'total_iterations': 5,
          'prefix': 'Total',
          'suffix': 'Writing',
          'decimals': 0,
          'bar_length': 30},
        { 'current_iteration': 509,
          'total_iterations': 1000,
          'prefix': '     ',
          'suffix': '/tmp3muil9sc/0000000000000000.tmp',
          'decimals': 1,
          'bar_length': 30},
    ]
    prepare_bars(config)
    print_bars(config)

    AUTHOR
    Christian Dönges, Platypus Projects GmbH www.platypus-projects.de

    The master repository for this file is located at
    https://rot.regenbogenkraut.de/cd/pymisclib.


    LICENSE
    This file is licensed under the Apache License, Version 2.0
    <https://spdx.org/licenses/Apache-2.0.html>.
    Please contact the author to discuss other licensing options.


    © Copyright 2019 Christian Dönges <cd.pymisclib@platypus-projects.net >

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import enum
import sys
from time import sleep


@enum.unique
class BAR_COLORS(enum.Enum):
    """Colors and the ANSI format strings used to create them."""
    Black = '\033[30m'
    Red = '\033[31m'
    Green = '\033[32m'
    Yellow = '\033[33m'
    Blue = '\033[34m'
    Magenta = '\033[35m'
    Cyan = '\033[36m'
    White = '\033[37m'
    BrightBlack = '\033[1;30m'
    BrightRed = '\033[1;31m'
    BrightGreen = '\033[1;32m'
    BrightYellow = '\033[1;33m'
    BrightBlue = '\033[1;34m'
    BrightMagenta = '\033[1;35m'
    BrightCyan = '\033[1;36m'
    BrightWhite = '\033[1;37m'


def print_single_bar(
        current_iteration: int,
        total_iterations: int,
        prefix: str = '',
        suffix: str = '',
        decimals: int = 1,
        bar_length: int = 80,
        bar_color: str = '\033[30m'):
    """Output a single progress bar to stdout.

    :param int current_iteration: the current iteration
    ;param int total_iterations: the total number of iterations
    :param str prefix: a string that will be output before the bar
    :param str suffix: a string that will be output after the bar
    :param int bar_length: the length of the bar in characters
    """
    current_iteration = min(current_iteration, total_iterations)

    percents = f'{100 * (current_iteration / float(total_iterations)):.{decimals}f}'
    filled_length = int(round(bar_length * current_iteration / float(total_iterations)))
    the_bar = f'{"█" * filled_length}{"-" * (bar_length - filled_length)}'

    sys.stdout.write(
        f'\x1b[2K\r{prefix} |{bar_color.value}{the_bar}'
        f'{BAR_COLORS.Black.value}| {percents}% {suffix}')


def prepare_bars(configs: list):
    """Print to prepare for the bars."""
    for c in configs:
        sys.stdout.write('\n')


def print_bars(configs: list):
    """Print progress bars."""
    # Move the cursor up to the start of the line of the first bar.
    for n in range(len(configs)):
        sys.stdout.write('\033[F')  # up

    for config in configs:
        ci = config['current_iteration']
        ti = config['total_iterations']
        prefix = config.get('prefix', '')
        suffix = config.get('suffix', '') + '\033[K\n'
        decimals = config.get('decimals', 1)
        bar_length = config.get('bar_length', 80)
        bar_color = config.get('bar_color', BAR_COLORS.Black)
        print_single_bar(ci, ti, prefix, suffix, decimals, bar_length,
                         bar_color)
    sys.stdout.flush()


def main():
    """Demonstration of how it works."""
    configs = [
        {
            'current_iteration': 0,
            'total_iterations': 5,
            'prefix': 'Total',
            'suffix': 'Writing',
            'decimals': 0,
            'bar_length': 30,
            'bar_color': BAR_COLORS.Green,
        },
        {
            'current_iteration': 509,
            'total_iterations': 1000,
            'prefix': '     ',
            'suffix': '/tmp3muil9sc/0000000000000000.tmp',
            'decimals': 1,
            'bar_length': 30
        },
        {
            'current_iteration': 1,
            'total_iterations': 10,
            'bar_length': 15,
            'bar_color': BAR_COLORS.BrightRed,
        },
        {
            'current_iteration': 0,
            'total_iterations': 100,
            'prefix': 'Jolly'
        },
    ]
    prepare_bars(configs)
    for i in range(100):
        print_bars(configs)
        for config in configs:
            config['current_iteration'] += 1
        sleep(.1)


if __name__ == '__main__':
    main()
